"""Icinga module tests."""
from datetime import timedelta
from unittest import mock

import pytest

from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet

from spicerack import icinga
from spicerack.administrative import Reason
from spicerack.remote import RemoteHosts


def set_mocked_icinga_host_output(mocked_icinga_host, output):
    """Setup the mocked icinga_host return value with the given output."""
    out = MsgTreeElem(output.encode(), parent=MsgTreeElem())
    mocked_icinga_host.run_sync.return_value = iter([(NodeSet('icinga-host'), out)])


class TestIcinga:
    """Test class for the Icinga class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.reason = Reason('Downtime reason', 'user1', 'orchestration-host', task_id='T12345')
        self.mocked_icinga_host = mock.MagicMock(spec_set=RemoteHosts)
        set_mocked_icinga_host_output(self.mocked_icinga_host, '/var/lib/icinga/rw/icinga.cmd')
        self.icinga = icinga.Icinga(self.mocked_icinga_host)

    def test_command_file_ok(self):
        """It should return the command_file setting from the Icinga configuration."""
        assert self.icinga.command_file == '/var/lib/icinga/rw/icinga.cmd'

    @pytest.mark.parametrize('output', (
        '',
        ' ',
    ))
    def test_command_file_raise(self, output):
        """It should raise IcingaError if failing to get the configuration value."""
        set_mocked_icinga_host_output(self.mocked_icinga_host, output)
        with pytest.raises(icinga.IcingaError, match='Unable to read command_file configuration'):
            self.icinga.command_file  # pylint: disable=pointless-statement

    def test_command_file_cached(self):
        """It should return the already cached value of the command_file if accessed again."""
        command_file = self.icinga.command_file
        self.mocked_icinga_host.reset_mock()
        assert self.icinga.command_file == command_file
        assert not self.mocked_icinga_host.called

    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_hosts_downtimed(self, mocked_time):
        """It should downtime the hosts on Icinga, yield and delete the downtime once done."""
        hosts = ['host1']
        call = 'icinga-downtime -h "host1" -d 14400 -r {reason}'.format(reason=self.reason.quoted())

        with self.icinga.hosts_downtimed(hosts, self.reason):
            self.mocked_icinga_host.run_sync.assert_called_once_with(call)
            self.mocked_icinga_host.run_sync.reset_mock()

        self.mocked_icinga_host.run_sync.assert_has_calls([
            mock.call('echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1" > /var/lib/icinga/rw/icinga.cmd')])
        assert mocked_time.called

    def test_downtime_is_kept_when_exception_is_raised(self):
        """Downtime should not be removed if an exception is raised."""
        hosts = ['host1']
        call = 'icinga-downtime -h "host1" -d 14400 -r {reason}'.format(reason=self.reason.quoted())

        with pytest.raises(ValueError):
            with self.icinga.hosts_downtimed(hosts, self.reason):
                self.mocked_icinga_host.run_sync.assert_called_once_with(call)
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert not self.mocked_icinga_host.run_sync.called

    def test_downtime_is_removed_when_exception_is_raised(self):
        """Downtime should not be removed if an exception is raised."""
        hosts = ['host1']
        call = 'icinga-downtime -h "host1" -d 14400 -r {reason}'.format(reason=self.reason.quoted())

        with pytest.raises(ValueError):
            with self.icinga.hosts_downtimed(hosts, self.reason, remove_on_error=True):
                self.mocked_icinga_host.run_sync.assert_called_once_with(call)
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert self.mocked_icinga_host.run_sync.called

    @pytest.mark.parametrize('hosts', (
        ['host1'],
        ['host1', 'host2'],
        NodeSet('host1'),
        NodeSet('host[1-9]'),
    ))
    def test_downtime_hosts_default_params(self, hosts):
        """It should downtime the hosts on the Icinga server with the default params."""
        self.icinga.downtime_hosts(hosts, self.reason)
        calls = [('icinga-downtime -h "{host}" -d 14400 -r {reason}').format(host=host, reason=self.reason.quoted())
                 for host in hosts]
        self.mocked_icinga_host.run_sync.assert_called_once_with(*calls)

    def test_downtime_hosts_custom_duration(self):
        """It should downtime the hosts for the given duration on the Icinga server."""
        self.icinga.downtime_hosts(['host1'], self.reason, duration=timedelta(minutes=30))
        self.mocked_icinga_host.run_sync.assert_called_once_with((
            'icinga-downtime -h "host1" -d 1800 -r {reason}'.format(reason=self.reason.quoted())))

    def test_downtime_hosts_invalid_duration(self):
        """It should raise IcingaError if the duration is too short."""
        with pytest.raises(icinga.IcingaError, match='Downtime duration must be at least 1 minute'):
            self.icinga.downtime_hosts(['host1'], self.reason, duration=timedelta(seconds=59))

    def test_downtime_hosts_no_hosts(self):
        """It should raise IcingaError if there are no hosts to downtime."""
        with pytest.raises(icinga.IcingaError, match='Got empty hosts list to downtime'):
            self.icinga.downtime_hosts([], self.reason)

    @pytest.mark.parametrize('hosts', (
        ['host1'],
        ['host1', 'host2'],
        NodeSet('host1'),
        NodeSet('host[1-9]'),
    ))
    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_host_command(self, mocked_time, hosts):
        """It should run the specified command for all the hosts on the Icinga server."""
        self.icinga.host_command('TEST_COMMAND', hosts, 'arg1', 'arg2')
        calls = [
            'echo -n "[1514764800] TEST_COMMAND;{host};arg1;arg2" > /var/lib/icinga/rw/icinga.cmd'.format(host=host)
            for host in hosts]

        self.mocked_icinga_host.run_sync.assert_called_with(*calls)
        assert mocked_time.called

    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_remove_downtime(self, mocked_time):
        """It should remove the downtime for the hosts on the Icinga server."""
        self.icinga.remove_downtime(NodeSet('host1'))
        self.mocked_icinga_host.run_sync.assert_called_with(
            'echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1" > /var/lib/icinga/rw/icinga.cmd')
        assert mocked_time.called



"""Confctl module tests."""
from unittest import mock

import pytest

from conftool.tests.unit import MockBackend

from spicerack import confctl
from spicerack.tests import get_fixture_path


class TestConfctl:
    """Confctl test class."""

    def setup_method(self):
        """Setup a Confctl instance with a mocked conftool backend and driver."""
        # pylint: disable=attribute-defined-outside-init
        self.conftool_backend = MockBackend({})
        confctl.kvobject.KVObject.backend = self.conftool_backend
        confctl.kvobject.KVObject.config = confctl.configuration.Config(driver='')
        config = get_fixture_path('confctl', 'config.yaml')
        schema = get_fixture_path('confctl', 'schema.yaml')
        with mock.patch('spicerack.confctl.kvobject.KVObject.setup'):
            self.confctl = confctl.Confctl(config=config, schema=schema, dry_run=False)
            self.entity = self.confctl._schema.entities['discovery']  # pylint: disable=protected-access

        self.entity.query = mock.MagicMock(return_value=[self.entity('test', 'dnsdisc')])
        self.discovery = confctl.ConftoolEntity(self.entity, dry_run=False)
        self.discovery_dry_run = confctl.ConftoolEntity(self.entity)

    def test_get_existing(self):
        """Calling get() should return the object matched by the tags."""
        for obj in self.discovery.get(dnsdisc='test'):
            assert obj.tags == {'dnsdisc': 'test'}

    def test_get_non_existing(self):
        """Calling get() without matches should raise ConfctlError."""
        self.entity.query.return_value = []
        with pytest.raises(confctl.ConfctlError, match='No match found'):
            list(self.discovery.get(dnsdisc='non-existing'))

    def test_update_ok(self):
        """Calling update() should update the objects matched by the tags."""
        self.discovery.update({'pooled': True}, dnsdisc='test')
        assert list(self.discovery.get(dnsdisc='test'))[0].pooled

    def test_update_dry_run(self):
        """Calling update() in dry_run mode should not update the objects matched by the tags."""
        list(self.discovery_dry_run.get(dnsdisc='test'))[0].update = mock.MagicMock(side_effect=Exception('test'))
        self.discovery_dry_run.update({'pooled': True}, dnsdisc='test')

    @pytest.mark.parametrize('exc_class, message', (
        (confctl.BackendError, 'Error writing to etcd'),
        (Exception, 'Generic error in conftool'),
    ))
    def test_update_errors(self, exc_class, message):
        """Calling update() should raise ConfctlError if there is an error in the backend."""
        list(self.discovery.get(dnsdisc='test'))[0].update = mock.MagicMock(side_effect=exc_class('test'))

        with pytest.raises(confctl.ConfctlError, match=message):
            self.discovery.update({'pooled': True}, dnsdisc='test')

    def test_set_and_verify_ok(self):
        """It should update the objects matched by the tags and check them."""
        self.discovery.set_and_verify('pooled', True, dnsdisc='test')
        assert list(self.discovery.get(dnsdisc='test'))[0].pooled

    def test_set_and_verify_fail(self):
        """It should raise ConfctlError if failing to check the modified objects."""
        list(self.discovery.get(dnsdisc='test'))[0].update = mock.MagicMock()  # Don't allow to update the record
        with pytest.raises(confctl.ConfctlError, match="Conftool key pooled has value 'False', expecting 'True'"):
            self.discovery.set_and_verify('pooled', True, dnsdisc='test')

    def test_set_and_verify_dry_run(self):
        """In dry_run mode it should not update the objects and not raise on failure to check them."""
        self.discovery_dry_run.set_and_verify('pooled', True, dnsdisc='test')
        assert not list(self.discovery.get(dnsdisc='test'))[0].pooled



"""Ganeti module"""

import logging

from typing import Dict, Optional

import requests
from requests.auth import HTTPBasicAuth

from spicerack.constants import PUPPET_CA_PATH
from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# This is the template for the SVC URL for the RAPI end point
CLUSTER_SVC_URL = 'https://ganeti01.svc.{dc}.wmnet:5030'
# These are the configured available set of rows by Ganeti cluster DC
CLUSTERS_AND_ROWS = {'eqiad': ('A', 'C'), 'codfw': ('A', 'B')}


class GanetiError(SpicerackError):
    """Raised on errors from Ganeti operations."""


class GanetiRAPI:
    """Class which wraps the read-only Ganeti RAPI"""

    def __init__(self, cluster_url: str, username: str, password: str, ca_path: str):
        """Initialize the instance

        Arguments:
            cluster (str): the short name of the cluster to access.
            username (str): the RAPI user name
            password (str): the RAPI user's password
            ca_path (str): the path to the signing certificate authority

        """
        self._auth = HTTPBasicAuth(username, password)
        self._ca_path = ca_path
        self._url = cluster_url

    def _api_get_request(self, *targets: str) -> Dict:
        """Perform a RAPI request.

        Arguments:
           *targets: The path components of the request (minus the /2/ part of the path)

        Returns:
           dict: The decoded JSON response

        Raises:
           spicerack.ganeti.GanetiError: on non-200 responses

        """
        full_url = '/'.join([self._url, '2'] + list(targets))
        result = requests.get(full_url, auth=self._auth, verify=self._ca_path)

        if result.status_code != 200:
            raise GanetiError('Non-200 from API: {}: {}'.format(result.status_code, result.text))

        return result.json()

    @property
    def info(self) -> Dict:
        """Return complete cluster information.

        Returns:
            dict: Cluster information dictionary

        Raises:
            spicerack.ganeti.GanetiError: API errors

        """
        return self._api_get_request('info')

    @property
    def master(self) -> Optional[str]:
        """Return the internal name for the current ganeti master node.

        Returns:
            str: The hostname of the master node (or None if the data is missing).

        Raises:
            spicerack.ganeti.GanetiError: API errors

        """
        return self.info.get('master')

    def fetch_instance(self, fqdn: str) -> Dict:
        """Return full information about an instance.

        Arguments:
           fqdn: the FQDN of the instance in question

        Returns:
           dict: host information

        Raises:
           spicerack.ganeti.GanetiError: API errors

        """
        return self._api_get_request('instances', fqdn)

    def fetch_instance_mac(self, fqdn: str) -> str:
        """Convenience method to return the 0th adapter's MAC address for an instance.

        Note that we don't allow the creation of instances with more than one MAC address at this time.

        Arguments:
           fqdn: the FQDN of the instance in question

        Returns:
           str: MAC address

        Raises:
           GanetiError: API errors

        """
        instance_info = self.fetch_instance(fqdn)
        if 'nic.macs' not in instance_info or not instance_info['nic.macs']:
            raise GanetiError("Can't find any MACs for instance")

        return instance_info['nic.macs'][0]


class Ganeti:
    """Class which wraps all Ganeti clusters."""

    def __init__(self, username: str, password: str):
        """Initialize the instance

        Arguments:
            username (str): The RAPI username to use.
            password (str): The RAPI password to use.

        """
        self._username = username
        self._password = password

    def rapi(self, cluster: str) -> GanetiRAPI:
        """Return a RAPI object for a particular cluster.

        Arguments:
            cluster (str): the name of the cluster to get a RAPI for.

        Returns:
            spicerack.ganeti.GanetiRAPI: the RAPI interface object

        Raises:
            spicerack.ganeti.GanetiError: on an invalid cluster name

        """
        if cluster not in CLUSTERS_AND_ROWS:
            raise GanetiError('Cannot find cluster {} (expected {}).'.format(cluster, tuple(CLUSTERS_AND_ROWS.keys())))

        cluster_url = CLUSTER_SVC_URL.format(dc=cluster)

        return GanetiRAPI(cluster_url, self._username, self._password, PUPPET_CA_PATH)



"""Ganeti Module test."""

import json

import pytest
import requests

from spicerack.exceptions import SpicerackError
from spicerack.ganeti import CLUSTER_SVC_URL, CLUSTERS_AND_ROWS, Ganeti, GanetiError, GanetiRAPI
from spicerack.tests import get_fixture_path, requests_mock_not_available


class TestGaneti:
    """Ganeti tests class"""

    def setup_method(self):
        """Setup test environment"""
        # pylint: disable=attribute-defined-outside-init

        self.ganeti = Ganeti(username='fake', password='password123')  # nosec

        self.cluster = 'eqiad'
        self.base_url = CLUSTER_SVC_URL.format(dc=self.cluster) + '/2'

        # load test fixtures
        with open(get_fixture_path('ganeti', 'info.json'), encoding='utf-8') as info_json:
            self.info = info_json.read()
        with open(get_fixture_path('ganeti', '404.json'), encoding='utf-8') as fourohfour_json:
            self.fourohfour = fourohfour_json.read()
        with open(get_fixture_path('ganeti', 'instance.json'), encoding='utf-8') as instance_json:
            self.instance_info = instance_json.read()
        with open(get_fixture_path('ganeti', 'bogus.json'), encoding='utf-8') as bogus_json:
            self.bogus_data = bogus_json.read()

    @pytest.mark.parametrize('cluster', CLUSTERS_AND_ROWS.keys())
    def test_ganeti_clusters(self, cluster):
        """Ganeti.rapi() should return a GanetiRAPI object."""
        assert isinstance(self.ganeti.rapi(cluster), GanetiRAPI)

    def test_ganeti_clusters_failure(self):
        """If a cluster that doesn't exist is passed to Ganeti.rapi(), it should except with a SpicerackError."""
        with pytest.raises(SpicerackError, match=r'Cannot find cluster bogus cluster \(expected .*'):
            self.ganeti.rapi('bogus cluster')

    @pytest.mark.skipif(requests_mock_not_available(), reason='Requires requests-mock fixture')
    def test_ganeti_rapi_master_success(self, requests_mock):
        """The master property of a RAPI should be the hostname for the master of this cluster."""
        requests_mock.get(self.base_url + '/info', text=self.info)

        master = json.loads(self.info)['master']
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master == master

    @pytest.mark.skipif(requests_mock_not_available(), reason='Requires requests-mock fixture')
    def test_ganeti_rapi_master_failure(self, requests_mock):
        """If a master is not specified by the uptsream API, the value of master on a RAPI should be None."""
        requests_mock.get(self.base_url + '/info', text=self.bogus_data)
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master is None

    @pytest.mark.skipif(requests_mock_not_available(), reason='Requires requests-mock fixture')
    def test_ganeti_rapi_instance_notfound(self, requests_mock):
        """A RAPI object should raise a GanetiError if a requested host does not exist."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(
            self.base_url + '/instances/testhost_404', text=self.fourohfour, status_code=requests.codes['not_found']
        )
        with pytest.raises(GanetiError, match=r'Non-200 from API: 404:.*'):
            rapi.fetch_instance('testhost_404')

    @pytest.mark.skipif(requests_mock_not_available(), reason='Requires requests-mock fixture')
    def test_ganeti_rapi_instance_invalid(self, requests_mock):
        """If no mac is present in host data, fetch_instance_mac should raise a GanetiError."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.base_url + '/instances/testhost_invalid', text=self.bogus_data)
        with pytest.raises(GanetiError, match=''):
            rapi.fetch_instance_mac('testhost_invalid')

    @pytest.mark.skipif(requests_mock_not_available(), reason='Requires requests-mock fixture')
    def test_ganeti_rapi_instance_valid(self, requests_mock):
        """The MAC returned by RAPI.fetch_instance_mac should match the data returned by the API."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.base_url + '/instances/testhost', text=self.instance_info)
        mac = json.loads(self.instance_info)['nic.macs'][0]
        assert mac == rapi.fetch_instance_mac('testhost')
